@@ -40,6 +40,10 @@ public:
 
   const char *connectionInfo();
 
+  unsigned getUpdateCount();
+  unsigned getPixelCount();
+  unsigned getPosition();
+
   // FdInStreamBlockCallback methods
   void blockCallback();
 
@@ -66,10 +70,12 @@ public:
   void dataRect(const rfb::Rect& r, int encoding);
 
   void setCursor(int width, int height, const rfb::Point& hotspot,
-                 void* data, void* mask);
+                 const rdr::U8* data);
 
   void fence(rdr::U32 flags, unsigned len, const char data[]);
 
+  void setLEDState(unsigned int state);
+
 private:
 
   void resizeFramebuffer();
@@ -89,6 +95,9 @@ private:
 
   DesktopWindow *desktop;
 
+  unsigned updateCount;
+  unsigned pixelCount;
+
   rfb::PixelFormat serverPF;
   rfb::PixelFormat fullColourPF;
 
